from tkinter.tix import IMMEDIATE
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from typing import Iterable, Optional, Tuple
import pandas as pd
from matplotlib.colors import CSS4_COLORS
from sklearn.manifold import TSNE as sklearn_TSNE


class TSNE:
    X_NAME = 'tsne-x'
    Y_NAME = 'tsne-y'
    RANDOM_SEED = 42
    
    def __init__(self, descriptors:np.array) -> None:
        self.__descriptors = np.copy(descriptors)
    
    @staticmethod
    def tsne(descriptors:np.array, seed=RANDOM_SEED) -> np.array:
        tsne = sklearn_TSNE(n_components=2, random_state=seed, learning_rate='auto', init='pca')
        return tsne.fit_transform(descriptors)
    
    @staticmethod
    def css_color_map(color_category:Iterable) -> dict:
        unique_colors = set(color_category)
        
        css_colors = list(CSS4_COLORS.keys())
        
        color_map = {}
        for i, color in enumerate(unique_colors):
            color_map[color] = css_colors[i]
        return color_map
    
    @staticmethod
    def tsne_df(descriptors, x_name:str=X_NAME, y_name:str=Y_NAME) -> pd.DataFrame:
        arr = TSNE.tsne(descriptors)
        return pd.DataFrame({
            x_name: arr[:, 0],
            y_name: arr[:, 1]
        })
    
    @staticmethod
    def plot(df:pd.DataFrame, x_col_name:str=X_NAME, y_col_name:str=Y_NAME, color_category:Optional[Iterable]=None,
             css_color_map:bool=False, opacity:float=1) -> go.Figure:
        color = None
        
        if color_category:
            df['color'] = color_category
            color = 'color'
            df['color'] = df['color'].fillna('missing color')  # Replace NaN w/ string bc px doesn't like NaN
        
        if css_color_map:
            plot = px.scatter(df, x=x_col_name, y=y_col_name, color=color, render_mode='svg', opacity=opacity,
                              color_discrete_map=TSNE.css_color_map(color_category))
        else:
            plot = px.scatter(df, x=x_col_name, y=y_col_name, color=color, render_mode='svg', opacity=opacity,
                              color_discrete_sequence=px.colors.qualitative.Alphabet)
        plot.update_layout(title={'x': 0.5})
        return plot
    
    def main(self, color_category:Optional[Iterable]=None, css_color_map:bool=False) -> Tuple[go.Figure, pd.DataFrame]:
        df = self.tsne_df(self.__descriptors)
        return (
            TSNE.plot(df, color_category=color_category, css_color_map=css_color_map),
            df
        )
        


from typing import Iterable, Optional, Tuple
import numpy as np
import plotly.graph_objects as go
import plotly.express as px
from rdkit import DataStructs


class SimMat:
    def __init__(self, row_prints:Iterable, col_prints:Iterable, key_type:str='ecfp',
                 row_labels:Optional[Iterable[str]]=None, col_labels:Optional[Iterable[str]]=None) -> None:
        self.__row_prints = row_prints
        self.__col_prints = col_prints
        self.__key_type = key_type
        self.__row_labels = row_labels
        self.__col_labels = col_labels
    
    @staticmethod
    def sim_matrix(row_prints:Iterable, col_prints:Iterable, key_type:str='ecfp') -> np.array:
        full_array = np.zeros((len(row_prints), len(col_prints)))

        for i, row in enumerate(row_prints):
            for j, col in enumerate(col_prints):
                if key_type == 'ecfp':
                    similarity = DataStructs.TanimotoSimilarity(row, col)
                elif key_type == 'maccs':
                    similarity = DataStructs.FingerprintSimilarity(row, col)
                else:
                    raise('Key type must be either "ecfp" or "maccs".')
                    
                full_array[i][j] = similarity
        
        return full_array
    
    @staticmethod
    def plot(sim_arr:np.array, row_labels:Optional[Iterable[str]]=None,
             col_labels:Optional[Iterable[str]]=None) -> go.Figure:
        fig = px.imshow(sim_arr, x=row_labels, y=col_labels)
        fig.update_layout(title={ 'x': 0.5 })
        return fig
    
    def main(self) -> Tuple[go.Figure, np.array]:
        sim_arr = self.sim_matrix(row_prints=self.__row_prints, col_prints=self.__col_prints, key_type=self.__key_type)
        fig = self.plot(sim_arr, row_labels=self.__row_labels, col_labels=self.__col_labels)
        return (fig, sim_arr)